from __future__ import annotations

from typing import Tuple

from PyQt5 import QtWidgets, QtCore, QtGui

from mado.input_base import InputBase
from mado.widgets.misc import make_icon_button


class TextInput(InputBase, QtWidgets.QLineEdit):
    def __init__(self, placeholder: str = None, text_edited: QtCore.pyqtSlot = None) -> None:
        super().__init__()

        self.setSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Preferred)

        if placeholder is not None:
            self.setPlaceholderText(placeholder)
        if text_edited is not None:
            self.textEdited.connect(text_edited)

    def set_value(self, value: any) -> None:
        self.setText(str(value))

    def get_value(self) -> str:
        return self.text()


class PasswordInput(InputBase, QtWidgets.QWidget):
    text_input: TextInput
    reveal_button: QtWidgets.QPushButton
    layout: QtWidgets.QHBoxLayout
    password_revealed: bool

    def __init__(self, placeholder: str = None, text_edited: QtCore.pyqtSlot = None) -> None:
        super().__init__()

        self.password_revealed = False

        self.text_input = TextInput(placeholder, text_edited)
        self.reveal_button = make_icon_button(icon=QtGui.QIcon(":/icons/view_record"))
        self.reveal_button.clicked.connect(self.on_reveal_button_clicked)
        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.text_input)
        self.layout.addWidget(self.reveal_button)

        self.setLayout(self.layout)
        self.layout.setContentsMargins(0, 0, 0, 0)

        self.text_input.setEchoMode(QtWidgets.QLineEdit.EchoMode.Password)

    def set_value(self, value: str) -> None:
        self.text_input.set_value(value)

    def get_value(self) -> str:
        return self.text_input.get_value()

    @QtCore.pyqtSlot()
    def on_reveal_button_clicked(self) -> None:
        self.password_revealed = not self.password_revealed

        if self.password_revealed:
            self.text_input.setEchoMode(QtWidgets.QLineEdit.EchoMode.Normal)
        else:
            self.text_input.setEchoMode(QtWidgets.QLineEdit.EchoMode.Password)


class MultilineTextInput(InputBase, QtWidgets.QTextEdit):
    def __init__(self, placeholder: str = None, text_edited: QtCore.pyqtSlot = None) -> None:
        super().__init__()

        self.setSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Preferred)

        if placeholder is not None:
            self.setPlaceholderText(placeholder)
        if text_edited is not None:
            self.textChanged.connect(text_edited)

    def set_value(self, value: any) -> None:
        self.setText(str(value))

    def get_value(self) -> str:
        return self.toPlainText()


class NumberInput(InputBase, QtWidgets.QSpinBox):
    def __init__(self, number_edited: QtCore.pyqtSlot = None, values_range: Tuple[int, int] = (0, 99)):
        super().__init__()

        self.setRange(*values_range)

        if number_edited:
            self.valueChanged.connect(number_edited)

    def set_value(self, value: int) -> None:
        self.setValue(value)

    def get_value(self) -> int:
        return self.value()



from abc import ABC, abstractmethod
from typing import Sequence, Dict, List

from sqlalchemy import Table, RowMapping, select, update, delete, insert

from mado.db import get_conn


class DataRepository(ABC):
    @abstractmethod
    def fetch_one(self, id_: int) -> RowMapping | Dict[str, any] | None:
        pass

    @abstractmethod
    def fetch_all(self) -> Sequence[RowMapping] | List[Dict[str, any]]:
        pass

    @abstractmethod
    def delete(self, id_: int) -> None:
        pass

    @abstractmethod
    def insert(self, data: Dict[str, any]) -> int:
        pass

    @abstractmethod
    def update(self, id_: int, data: Dict[str, any]) -> None:
        pass


class BasicDataRepository(DataRepository):
    table: Table

    def __init__(self, table: Table) -> None:
        self.table = table

    def fetch_one(self, id_: int) -> RowMapping | None:
        return get_conn().execute(select(self.table).where(self.table.c.id == id_)).mappings().first()

    def fetch_all(self) -> Sequence[RowMapping]:
        return get_conn().execute(select(self.table)).mappings().all()

    def delete(self, id_: int) -> None:
        get_conn().execute(delete(self.table).where(self.table.c.id == id_))
        get_conn().commit()

    def insert(self, data: Dict[str, any]) -> int:
        res = get_conn().execute(insert(self.table).values(data))
        get_conn().commit()

        return res.lastrowid

    def update(self, id_: int, data: Dict[str, any]) -> None:
        get_conn().execute(update(self.table).where(self.table.c.id == id_).values(data))
        get_conn().commit()



from typing import Type, Callable, List

from mado.input_base import InputBase
from PyQt5 import QtWidgets, QtCore, QtGui

from mado.widgets.misc import make_button


class ListBody(QtWidgets.QListWidget):

    def __init__(self) -> None:
        super().__init__()

        self.setSelectionMode(QtWidgets.QListWidget.SelectionMode.SingleSelection)

    def get_value(self) -> List[any]:
        res = list()

        for i in range(self.count()):
            list_item_widget = self.itemWidget(self.item(i))
            inp = list_item_widget.layout().itemAt(0).widget()

            res.append(inp.get_value())

        return res

    def add_entry(self, inp: InputBase | QtWidgets.QWidget) -> InputBase:
        item = QtWidgets.QListWidgetItem()

        w = QtWidgets.QWidget()
        w.setLayout(QtWidgets.QVBoxLayout())
        w.layout().setContentsMargins(10, 20, 10, 20)
        w.layout().addWidget(inp)

        item.setSizeHint(w.sizeHint())

        self.addItem(item)
        self.setItemWidget(item, w)

        return inp


class ListInput(InputBase, QtWidgets.QWidget):
    list_body: ListBody

    control_box: QtWidgets.QHBoxLayout
    layout: QtWidgets.QVBoxLayout

    add_button: QtWidgets.QPushButton
    delete_button: QtWidgets.QPushButton

    input_delegate: Type[InputBase] | Callable[[], InputBase]

    def __init__(self, input_delegate: Type[InputBase] | Callable[[], InputBase],
                 list_body: ListBody = None) -> None:
        super().__init__()

        if list_body is None:
            list_body = ListBody()

        self.list_body = list_body
        self.input_delegate = input_delegate

        self.add_button = make_button("Добавить", icon=QtGui.QIcon(":/icons/new_record"))
        self.delete_button = make_button("Удалить", icon=QtGui.QIcon(":/icons/delete_record"))
        self.delete_button.setEnabled(False)
        self.control_box = QtWidgets.QHBoxLayout()
        self.control_box.addWidget(self.add_button)
        self.control_box.addStretch()
        self.control_box.addWidget(self.delete_button)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addLayout(self.control_box)
        self.layout.addWidget(self.list_body)

        self.setLayout(self.layout)

        self.list_body.itemClicked.connect(lambda: self.delete_button.setEnabled(True))
        self.list_body.model().modelReset.connect(lambda: self.delete_button.setEnabled(False))

        self.add_button.clicked.connect(self.on_add_button_clicked)
        self.delete_button.clicked.connect(self.on_delete_button_clicked)

    def get_value(self) -> List[any]:
        return self.list_body.get_value()

    def set_value(self, value: List[any]) -> None:
        self.list_body.clear()

        for x in value:
            inp = self.list_body.add_entry(self.input_delegate())

            inp.set_value(x)

    @QtCore.pyqtSlot()
    def on_add_button_clicked(self) -> None:
        self.list_body.add_entry(self.input_delegate())

    @QtCore.pyqtSlot()
    def on_delete_button_clicked(self) -> None:
        row = self.list_body.currentRow()

        if row != -1:
            self.list_body.takeItem(row)



from __future__ import annotations

from PyQt5 import QtWidgets, QtCore

from mado.input_base import InputBase

from datetime import date, datetime, time


class DateInput(InputBase, QtWidgets.QDateEdit):
    def __init__(self, date_changed: QtCore.pyqtSlot = None) -> None:
        super().__init__()

        self.setCalendarPopup(True)
        self.setDate(QtCore.QDate.currentDate())

        if date_changed is not None:
            self.dateChanged.connect(lambda x: date_changed(x.toPyDate()))

    def get_value(self) -> date:
        return self.date().toPyDate()

    def set_value(self, value: date) -> None:
        self.setDate(QtCore.QDate(value.year, value.month, value.day))

    def set_readonly(self) -> None:
        self.setButtonSymbols(QtWidgets.QDateEdit.ButtonSymbols.NoButtons)
        self.setCalendarPopup(False)


class DateTimeInput(InputBase, QtWidgets.QDateTimeEdit):
    def __init__(self, datetime_changed: QtCore.pyqtSlot = None) -> None:
        super().__init__()

        self.setCalendarPopup(True)
        self.setDateTime(QtCore.QDateTime.currentDateTime())

        if datetime_changed is not None:
            self.dateTimeChanged.connect(lambda x: datetime_changed(x.toPyDateTime()))

    def get_value(self) -> datetime:
        return self.dateTime().toPyDateTime()

    def set_value(self, value: datetime) -> None:
        self.setDateTime(QtCore.QDateTime(value.year, value.month, value.day, value.hour, value.minute))

    def set_readonly(self) -> None:
        self.setCalendarPopup(False)


class TimeInput(InputBase, QtWidgets.QTimeEdit):
    def __init__(self, time_changed: QtCore.pyqtSlot = None) -> None:
        super().__init__()

        self.setTime(QtCore.QTime.currentTime())

        if time_changed is not None:
            self.timeChanged.connect(lambda x: time_changed(x.toPyTime()))

    def get_value(self) -> time:
        return self.time().toPyTime()

    def set_value(self, value: time) -> None:
        self.setTime(QtCore.QTime(value.hour, value.minute))



import sqlite3

from sqlalchemy import create_engine, Engine, MetaData, Connection, text, event

from mado.utils import absolute_appdata_path

engine: Engine = None
conn: Connection = None


def setup_database_connection(meta: MetaData) -> None:
    global engine, conn

    engine = create_engine(
        "sqlite:///" + absolute_appdata_path("storage.db"),
        echo=True,
    )

    raw = engine.raw_connection()
    for line in raw.iterdump():
        print(line)

    @event.listens_for(engine, "connect")
    def setup_functions(dbapi_connection: sqlite3.Connection, _) -> None:
        dbapi_connection.create_function("pylower", 1, str.lower)
        dbapi_connection.create_function("pyupper", 1, str.upper)

    meta.create_all(bind=engine)

    conn = engine.connect()
    conn.execute(text("PRAGMA foreign_keys = ON"))  # pk functionality have to be enabled manually


def get_conn() -> Connection:
    return conn



from enum import Enum
from typing import List, Tuple, Type

from PyQt5 import QtWidgets, QtCore, QtGui

from mado.input_base import InputBase


# TODO test
class RadioInput(InputBase, QtWidgets.QWidget):
    radio_selected = QtCore.pyqtSignal(object)
    selected_data: any

    buttons_group: QtWidgets.QButtonGroup
    layout: QtWidgets.QHBoxLayout

    @QtCore.pyqtSlot()
    def on_button_toggled(self) -> None:
        button: QtWidgets.QRadioButton = self.sender()

        if button.isChecked():
            self.selected_data = button.property("user_data")

            self.radio_selected.emit(self.selected_data)

    def __init__(self, variants: List[Tuple[any, str]], radio_selected: QtCore.pyqtSlot = None,
                 horizontal: bool = True) -> None:
        super().__init__()

        self.buttons_group = QtWidgets.QButtonGroup()
        self.layout = QtWidgets.QHBoxLayout() if horizontal else QtWidgets.QVBoxLayout()
        self.selected_data = None

        for user_data, text in variants:
            button = QtWidgets.QRadioButton()
            button.setText(text)
            button.setProperty("user_data", user_data)

            button.toggled.connect(self.on_button_toggled)

            self.buttons_group.addButton(button)
            self.layout.addWidget(button)
        self.layout.addStretch()

        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        if radio_selected is not None:
            self.radio_selected.connect(radio_selected)

    def get_value(self) -> any:
        return self.selected_data

    def set_value(self, value: any) -> None:
        for x in self.buttons_group.buttons():
            x: QtWidgets.QRadioButton

            if x.property("user_data") == value:
                x.setChecked(True)


# TODO test
class EnumRadioInput(RadioInput):
    enum: Type[Enum]

    def __init__(self, enum: Type[Enum], radio_selected: QtCore.pyqtSlot = None,
                 horizontal: bool = True) -> None:
        self.enum = enum
        to_list = [(x.name, x.value) for x in self.enum]

        super().__init__(to_list, radio_selected, horizontal)

    def get_value(self) -> any:
        key = self.selected_data
        if key is None:
            return None
        
        return self.enum[self.selected_data]

    def set_value(self, value: any) -> None:
        for x in self.buttons_group.buttons():
            x: QtWidgets.QRadioButton

            if x.property("user_data") == value.name:
                x.setChecked(True)


class BoolInput(InputBase, QtWidgets.QCheckBox):
    mark_changed = QtCore.pyqtSignal(bool)

    def __init__(self, label: str = None, mark_changed: QtCore.pyqtSlot = None):
        super().__init__()

        if label:
            self.setText(label)

        if mark_changed:
            self.mark_changed.connect(mark_changed)

        self.stateChanged.connect(lambda state: self.mark_changed.emit(state == QtCore.Qt.CheckState.Checked))

        self.setSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)

    def get_value(self) -> bool:
        return self.checkState() == QtCore.Qt.CheckState.Checked

    def set_value(self, value: bool) -> None:
        if value:
            self.toggle()



from typing import List

from PyQt5 import QtWidgets, QtGui, QtCore

from mado.input_base import InputBase


class TableInput(InputBase, QtWidgets.QTableView):
    model: QtGui.QStandardItemModel
    cell_edited = QtCore.pyqtSignal(tuple, str)

    def __init__(self, cell_edited: QtCore.pyqtSlot = None):
        super().__init__()

        self.model = QtGui.QStandardItemModel()
        self.setModel(self.model)

        self.model.itemChanged.connect(self.on_item_changed)

        if cell_edited is not None:
            self.cell_edited.connect(cell_edited)

    def set_value(self, value: List[List[str]]):
        for y, row in enumerate(value):
            for x, column in enumerate(row):
                if column is None:
                    continue

                self.model.setItem(y, x, QtGui.QStandardItem(column))

    def get_value(self) -> List[List[str]]:
        res = list()

        for row in range(self.model.rowCount()):
            row_data = list()
            for col in range(self.model.columnCount()):
                index = self.model.index(row, col)
                item = self.model.data(index)
                row_data.append(item)
            res.append(row_data)

        return res

    # здесь лучше не аннотировать pyqtSlot....
    def on_item_changed(self, item: QtGui.QStandardItem) -> None:
        self.cell_edited.emit((item.row(), item.column()), item.text())

    def reset_with_dimensions(self, rows: int, cols: int) -> None:
        self.model.clear()

        for _ in range(rows):
            self.model.appendRow([QtGui.QStandardItem() for _ in range(cols)])


class DelegatedTableInput(InputBase, QtWidgets.QTableView):
    model: QtGui.QStandardItemModel

    delegated_cells: dict[int, dict[int, InputBase]]

    def __init__(self):
        super().__init__()

        self.model = QtGui.QStandardItemModel()
        self.setModel(self.model)

        self.delegated_cells = dict()

    def set_value(self, value: List[List[any]]):
        for y, row in enumerate(value):
            for x, column in enumerate(row):
                if column is None:
                    continue

                delegated_cell = self.delegated_cells.get(y, dict()).get(x, None)
                if delegated_cell is not None:
                    delegated_cell.set_value(column)
                else:
                    self.model.setItem(y, x, QtGui.QStandardItem(column))

    def get_value(self) -> List[List[any]]:
        res = list()

        for row in range(self.model.rowCount()):
            row_data = list()
            for col in range(self.model.columnCount()):
                delegated_cell = self.delegated_cells.get(row, dict()).get(col, None)
                if delegated_cell is not None:
                    row_data.append(delegated_cell.get_value())
                else:
                    row_data.append(self.model.item(row, col).text())
            res.append(row_data)

        return res

    def set_custom_input(self, row: int, col: int, inp: InputBase | QtWidgets.QWidget) -> None:
        if self.delegated_cells.get(row) is None:
            self.delegated_cells[row] = dict()

        self.delegated_cells[row][col] = inp
        self.setIndexWidget(self.model.index(row, col), inp)

    def reset_with_dimensions(self, rows: int, cols: int) -> None:
        self.model.clear()

        for _ in range(rows):
            self.model.appendRow([QtGui.QStandardItem() for _ in range(cols)])



from __future__ import annotations

from PyQt5 import QtWidgets, QtGui, QtCore

from mado.application_window import WindowWidget


def make_font(family: str = None, size: int = None, bold: bool = False, italic: bool = False,
              underline: bool = False) -> QtGui.QFont:
    w = QtGui.QFont()

    if family is not None:
        w.setFamily(family)
    if size is not None:
        w.setPointSize(size)

    w.setBold(bold)
    w.setItalic(italic)
    w.setUnderline(underline)

    return w


def make_label(text: str = "", font: QtGui.QFont = None) -> QtWidgets.QLabel:
    w = QtWidgets.QLabel()

    w.setText(text)
    if font is not None:
        w.setFont(font)

    return w


def make_button(label: str = "", font: QtGui.QFont = None, flat: bool = False,
                icon: QtGui.QIcon = None, icon_size: int = 20) -> QtWidgets.QPushButton:
    w = QtWidgets.QPushButton()

    w.setText(label)
    if font is not None:
        w.setFont(font)
    if icon is not None:
        w.setIcon(icon)
        w.setIconSize(QtCore.QSize(icon_size, icon_size))

    w.setFlat(flat)

    return w


def make_icon_button(flat: bool = False, icon: QtGui.QIcon = None,
                     icon_size: int = 20) -> QtWidgets.QPushButton:
    w = QtWidgets.QPushButton()

    if icon is not None:
        w.setIcon(icon)
        w.setIconSize(QtCore.QSize(icon_size, icon_size))

    w.setFlat(flat)
    w.setSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)

    return w


def make_back_button(window: WindowWidget) -> QtWidgets.QPushButton:
    w = make_button("Назад", flat=True, icon=QtGui.QIcon(":/icons/go_back"), icon_size=20,
                    font=make_font(size=14))
    w.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)
    w.clicked.connect(window.pop_screen)

    return w



from __future__ import annotations

from PyQt5 import QtWidgets, QtCore

from mado.application_window import WindowWidget
from mado.entity.config import EntityConfig
from mado.screen import Screen
from mado.widgets.misc import make_font, make_label, make_back_button
from mado.widgets.tables import MRecordTable, fixed_hover_effects


class EntityExplorer(Screen, QtWidgets.QWidget):
    table: MRecordTable
    config: EntityConfig

    top_box: QtWidgets.QHBoxLayout
    control_box: QtWidgets.QHBoxLayout
    table_box: QtWidgets.QVBoxLayout
    layout: QtWidgets.QVBoxLayout

    back_button: QtWidgets.QPushButton
    name_label: QtWidgets.QLabel

    def __init__(self, window: WindowWidget, config: EntityConfig):
        super().__init__()

        self.config = config
        window.setWindowTitle(f"{self.config.entity_label}: обзор записей")

        self.make_ui(window)

        self.table.row_selected.connect(self.on_table_row_selected)
        self.table.model.modelReset.connect(self.on_table_model_reset)

        self.fill_table()

    def make_ui(self, window: WindowWidget) -> None:
        self.table = MRecordTable()
        self.table.model.setHorizontalHeaderLabels(self.config.table_labels)
        # self.table.setStyleSheet(fixed_hover_effects)

        self.name_label = make_label(self.config.entity_label, font=make_font(size=14))

        self.back_button = make_back_button(window)
        self.back_button.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)
        self.top_box = QtWidgets.QHBoxLayout()
        self.top_box.addWidget(self.back_button)
        self.top_box.addSpacing(5)
        self.top_box.addWidget(self.name_label)
        self.top_box.addStretch()

        self.control_box = QtWidgets.QHBoxLayout()

        self.table_box = QtWidgets.QVBoxLayout()
        self.table_box.addWidget(self.table)

        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        self.layout.addLayout(self.top_box)
        self.layout.addSpacing(10)
        self.layout.addLayout(self.control_box)
        self.layout.addLayout(self.table_box)

    def fill_table(self):
        self.table.set_data(
            [([x[y] for y in self.config.table_columns], x.id) for x in self.config.repository.fetch_all()])
        self.table.normalize()

    @QtCore.pyqtSlot()
    def on_table_row_selected(self) -> None:
        pass

    @QtCore.pyqtSlot()
    def on_table_model_reset(self) -> None:
        pass



from __future__ import annotations

from typing import List, Tuple, Optional

from PyQt5 import QtWidgets, QtGui, QtCore

fixed_hover_effects = """
QTableView::item:hover {
    background-color: transparent;
    margin: 0;
    padding: 0;
}

QTableView::item:selected {
    background-color: palette(hightlight);
    color: palette(highlightedText);
};"""


def make_item_with_data(value: str, item_data: any) -> QtGui.QStandardItem:
    item = QtGui.QStandardItem(str(value))
    item.setData(item_data)

    return item


class MReportTable(QtWidgets.QTableView):
    model: QtGui.QStandardItemModel

    def __init__(self):
        super().__init__()

        self.model = QtGui.QStandardItemModel()
        self.setModel(self.model)

        self.setEditTriggers(QtWidgets.QTableView.EditTrigger.NoEditTriggers)
        self.setSelectionMode(QtWidgets.QTableView.SelectionMode.NoSelection)

    def set_data(self, data: List[List[object]]) -> None:
        self.model.clear()

        for row in data:
            self.model.appendRow([QtGui.QStandardItem(str(col)) for col in row])


class MSelectableTable(QtWidgets.QTableView):
    model: QtGui.QStandardItemModel

    def __init__(self) -> None:
        super().__init__()

        self.model = QtGui.QStandardItemModel()
        self.setModel(self.model)

    def _selected_indexes_to_cells(self, indexes: List[QtCore.QModelIndex]) -> List[Tuple[Tuple[int, int], any]]:
        cells = [((x.row(), x.column()), self.model.itemFromIndex(x).data())
                 for x in indexes]

        return cells

    def get_selected_cells(self) -> List[Tuple[Tuple[int, int], any]]:
        selection_model = self.selectionModel()
        selected_indexes = selection_model.selectedIndexes()

        return self._selected_indexes_to_cells(selected_indexes)


class MRecordTable(MSelectableTable):
    row_double_clicked = QtCore.pyqtSignal(object)
    row_selected = QtCore.pyqtSignal(tuple)

    def __init__(self) -> None:
        super().__init__()

        self.setEditTriggers(QtWidgets.QTableView.EditTrigger.NoEditTriggers)
        self.setSelectionMode(QtWidgets.QTableView.SelectionMode.SingleSelection)
        self.setSelectionBehavior(QtWidgets.QTableView.SelectionBehavior.SelectRows)
        self.verticalHeader().setDefaultSectionSize(45)

        self.model.setHorizontalHeaderLabels(["Наименование"])

        self.doubleClicked.connect(
            lambda index: self.row_double_clicked.emit(self.model.itemFromIndex(index).data()))
        self.selectionModel().selectionChanged.connect(
            lambda selection: self.row_selected.emit(self._selected_indexes_to_cells(selection.indexes())[0]))

    def normalize(self):
        if self.model.columnCount() > 0:
            self.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.Stretch)

    def set_data(self, data: List[Tuple[List[any], any]]):
        horizontal_labels = [self.model.horizontalHeaderItem(x).text() for x in range(self.model.columnCount())]

        self.model.clear()
        self.model.setHorizontalHeaderLabels(horizontal_labels)

        for x in data:
            rows, payload = x

            self.model.appendRow([make_item_with_data(x, payload) for x in rows])

    def get_selected_row(self) -> Optional[any]:
        cells = self.get_selected_cells()

        return cells[0][1] if len(cells) > 0 else None



from __future__ import annotations

from typing import Dict

from PyQt5 import QtWidgets, QtCore
from sqlalchemy import exc

from mado.entity.entity_explorer import EntityExplorer
from mado.utils import make_warning
from mado.widgets.modal import Modal


class WithCreateOperation:
    create_button: QtWidgets.QPushButton

    @QtCore.pyqtSlot()
    def open_create_screen(self: EntityExplorer) -> None:
        modal = Modal(self)

        creator = self.config.creator(modal, self.config)

        created_id = -1
        created_data = dict()

        @QtCore.pyqtSlot(int, dict)
        def on_record_created(id_: int, data: Dict[str, any]) -> None:
            nonlocal created_id, created_data

            created_id = id_
            created_data = data

            self.fill_table()

        @QtCore.pyqtSlot()
        def on_record_creation_finalized() -> None:
            manager = self.config.manager(modal, self.config, created_id, created_data)

            manager.record_edited.connect(lambda *args, **kwargs: self.fill_table())

            modal.push_anon_screen(manager)

        creator.record_created.connect(on_record_created)
        creator.record_creation_finalized.connect(on_record_creation_finalized)

        modal.change_anon_screen(creator)
        modal.exec()


class WithManageOperation:
    manage_button: QtWidgets.QPushButton

    @QtCore.pyqtSlot()
    def open_manage_screen(self: EntityExplorer):
        modal = Modal(self)

        row = self.table.get_selected_row()
        if row is None:
            return

        manager = self.config.manager(modal, self.config, row)
        manager.record_edited.connect(lambda *args, **kwargs: self.fill_table())

        modal.change_anon_screen(manager)
        modal.exec()


class WithDeleteOperation:
    delete_button: QtWidgets.QPushButton

    @QtCore.pyqtSlot()
    def on_delete_button_clicked(self) -> None:
        id_ = self.table.get_selected_row()
        if id_ is None:
            return

        dialog = QtWidgets.QMessageBox()
        dialog.setIcon(QtWidgets.QMessageBox.Icon.Question)
        dialog.setWindowTitle(f"{self.config.entity_label}: удаление записи")
        dialog.setText("Вы действительно хотите удалить запись?")
        dialog.setStandardButtons(QtWidgets.QMessageBox.StandardButton.No | QtWidgets.QMessageBox.StandardButton.Yes)
        dialog.button(QtWidgets.QMessageBox.StandardButton.Yes).setText("Да")
        dialog.button(QtWidgets.QMessageBox.StandardButton.No).setText("Нет")

        reply = dialog.exec()
        if reply == QtWidgets.QMessageBox.StandardButton.No:
            return

        try:
            self.config.repository.delete(id_)
        except exc.IntegrityError as e:
            make_warning(self, self.config.process_delete_error(str(e.orig)), "Ошибка").exec()

        self.fill_table()


class WithViewOperation:
    view_button: QtWidgets.QPushButton

    @QtCore.pyqtSlot()
    def open_view_screen(self) -> None:
        modal = Modal(self)

        row = self.table.get_selected_row()
        if row is None:
            return

        viewer = self.config.viewer(modal, self.config, lambda *args, **kwargs: modal.accept(), row)

        modal.change_anon_screen(viewer)
        modal.exec()



from __future__ import annotations

from enum import Enum
from typing import Dict, OrderedDict

from PyQt5 import QtWidgets, QtCore

from mado.input_base import InputBase
from mado.utils import make_warning
from mado.capabilities import ReadonlyCapable


class FormBody(QtWidgets.QWidget):
    _readonly: bool
    _drawn: bool

    widgets: OrderedDict[str, InputBase | QtWidgets.QWidget]
    _labels: Dict[str, str]

    def __init__(self):
        super().__init__()

        self.widgets = OrderedDict[str, InputBase | QtWidgets.QWidget]()
        self._labels = dict()
        self._readonly = False
        self._drawn = False

        self.layout = QtWidgets.QFormLayout()
        self.setLayout(self.layout)

    def add(self, id_: str, w: QtWidgets.QWidget, label: str = None) -> FormBody:
        if label is not None:
            self._labels[id_] = label

        self.widgets[id_] = w

        return self

    def draw(self) -> FormBody:
        if self._drawn:
            raise Exception("already drawn")
        self._drawn = True

        if self._readonly:
            self.setAttribute(QtCore.Qt.WidgetAttribute.WA_TransparentForMouseEvents)
            self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

        for k, v in self.widgets.items():
            if self._readonly:
                if isinstance(v, ReadonlyCapable):
                    v.set_readonly()

                v.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

            if self._labels.get(k, None):
                self.layout.addRow(self._labels[k], v)
            else:
                self.layout.addRow(v)

        return self

    def set_value(self, value: Dict[str, any]) -> FormBody:
        for k, v in value.items():
            if self.widgets.get(k, None) is not None:
                self.widgets[k].set_value(v)

        return self

    def get_value(self) -> Dict[str, any]:
        res = dict()

        for k, v in self.widgets.items():
            if isinstance(v, InputBase):
                res[k] = v.get_value()

        return res

    def warn(self, msg: str) -> None:
        make_warning(self, msg).exec()

    def set_readonly(self) -> FormBody:
        self._readonly = True

        return self


class FormType(Enum):
    VIEW = 1
    MANAGE = 2
    CREATE = 3



from __future__ import annotations
from typing import Dict, Type, Callable, TYPE_CHECKING

from PyQt5 import QtWidgets, QtCore

from mado.application_window import Window

if TYPE_CHECKING:
    from mado.application_window import WindowWidget
    from mado.screen import ScreenWidget


class Modal(Window, QtWidgets.QDialog):
    screens: Dict[str, Type[ScreenWidget] | Callable[[WindowWidget, ...], ScreenWidget]]
    stack: QtWidgets.QStackedLayout

    def __init__(self, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent)

        self.stack = QtWidgets.QStackedLayout()
        self.screens = dict()

        size = self.screen().size()
        self.resize(int(size.width() * 0.55), int(size.height() * 0.7))

        self.stack = QtWidgets.QStackedLayout()

        self.setLayout(self.stack)

    def register_screen(self, id_: str,
                        screen: Type[ScreenWidget] | Callable[[WindowWidget, ...], ScreenWidget]) -> None:
        self.screens[id_] = screen

    def change_anon_screen(self, screen: ScreenWidget) -> None:
        self.push_anon_screen(screen)

        while self.stack.count() > 1:
            widget = self.stack.takeAt(0).widget()
            if widget:
                widget.deleteLater()

    def push_anon_screen(self, screen: ScreenWidget) -> None:
        self.stack.addWidget(screen)

        self.stack.setCurrentIndex(self.stack.count() - 1)

    def change_screen(self, id_: str, *args, **kwargs) -> None:
        self.change_anon_screen(self.screens[id_](self, *args, **kwargs))

    def push_screen(self, id_: str, *args, **kwargs) -> None:
        self.push_anon_screen(self.screens[id_](self, *args, **kwargs))

    @QtCore.pyqtSlot()
    def pop_screen(self) -> None:
        widget = self.stack.widget(self.stack.count() - 1)
        self.stack.removeWidget(widget)
        widget.deleteLater()

        screen: ScreenWidget = self.stack.widget(self.stack.count() - 1)
        screen.reconfigure()



from __future__ import annotations

from PyQt5 import QtCore, QtGui

from mado.application_window import WindowWidget
from mado.entity.config import EntityConfig
from mado.entity.entity_explorer import EntityExplorer
from mado.entity.explorer_mixins import WithCreateOperation, WithDeleteOperation, WithManageOperation, WithViewOperation
from mado.widgets.misc import make_button


class ReadonlyDirectoryExplorer(EntityExplorer, WithViewOperation):
    def __init__(self, window: WindowWidget, config: EntityConfig) -> None:
        super().__init__(window, config)

        self.view_button.clicked.connect(self.open_view_screen)
        self.table.row_double_clicked.connect(self.open_view_screen)

    def make_ui(self, window: WindowWidget) -> None:
        super().make_ui(window)

        self.view_button = make_button("Просмотреть", icon=QtGui.QIcon(":/icons/view_record"))
        self.view_button.setEnabled(False)

        self.control_box.addStretch()
        self.control_box.addWidget(self.view_button)

    @QtCore.pyqtSlot()
    def on_table_row_selected(self) -> None:
        self.view_button.setEnabled(True)

    @QtCore.pyqtSlot()
    def on_table_model_reset(self) -> None:
        self.view_button.setEnabled(False)


class DirectoryExplorer(EntityExplorer, WithCreateOperation, WithDeleteOperation, WithManageOperation):
    def __init__(self, window: WindowWidget, config: EntityConfig):
        super().__init__(window, config)

        self.table.row_double_clicked.connect(self.open_manage_screen)

        self.manage_button.clicked.connect(self.open_manage_screen)
        self.create_button.clicked.connect(self.open_create_screen)
        self.delete_button.clicked.connect(self.on_delete_button_clicked)

    def make_ui(self, window: WindowWidget) -> None:
        super().make_ui(window)

        self.manage_button = make_button("Открыть запись")
        # icon=QtGui.QIcon(":/icons/manage_record")
        # self.manage_button.setIconSize(QtCore.QSize(30, 20))
        self.manage_button.setEnabled(False)

        self.delete_button = make_button("Удалить", icon=QtGui.QIcon(":/icons/delete_record"))
        self.delete_button.setEnabled(False)

        self.create_button = make_button("Создать", icon=QtGui.QIcon(":/icons/new_record"))

        self.control_box.addWidget(self.create_button)
        self.control_box.addWidget(self.delete_button)
        self.control_box.addStretch()
        self.control_box.addWidget(self.manage_button)

    @QtCore.pyqtSlot()
    def on_table_row_selected(self) -> None:
        self.manage_button.setEnabled(True)
        self.delete_button.setEnabled(True)

    @QtCore.pyqtSlot()
    def on_table_model_reset(self) -> None:
        self.manage_button.setEnabled(False)
        self.delete_button.setEnabled(False)



from __future__ import annotations

from abc import abstractmethod
from typing import Callable, TYPE_CHECKING

from PyQt5 import QtWidgets, QtCore, QtGui
from sqlalchemy import exc

from mado.application_window import Window, WindowWidget
from mado.form_body import FormBody, FormType
from mado.screen import Screen
from mado.utils import make_warning, make_info
from mado.widgets.misc import make_button

if TYPE_CHECKING:
    from mado.entity.config import EntityConfig


class RecordCreator(Screen):
    record_created: QtCore.pyqtSignal
    record_creation_finalized: QtCore.pyqtSignal

    config: EntityConfig

    @abstractmethod
    def __init__(self, window: WindowWidget, config: EntityConfig) -> None:
        super().__init__()


class RecordManager(Screen):
    record_edited: QtCore.pyqtSignal
    record_edition_finalized: QtCore.pyqtSignal

    config: EntityConfig

    @abstractmethod
    def __init__(self, window: WindowWidget, config: EntityConfig, id_: int,
                 initial_data: dict = None) -> None:
        super().__init__()


class RecordViewer(Screen):
    close_viewer: callable
    config: EntityConfig

    @abstractmethod
    def __init__(self, window: WindowWidget, config: EntityConfig, close_viewer: callable, id_: int,
                 initial_data: dict = None) -> None:
        super().__init__()


class BasicRecordCreator(RecordCreator, QtWidgets.QWidget):
    record_created = QtCore.pyqtSignal(int, dict)
    record_creation_finalized = QtCore.pyqtSignal()

    control_box: QtWidgets.QHBoxLayout
    form_box: QtWidgets.QVBoxLayout
    create_button: QtWidgets.QPushButton
    layout: QtWidgets.QVBoxLayout

    form_body: FormBody

    def __init__(self, window: WindowWidget, config: EntityConfig) -> None:
        super().__init__(window, config)

        self.config = config
        self.form_body = self.config.make_form(FormType.CREATE)

        self.make_ui()

        self.create_button.clicked.connect(self.on_create_button_clicked)

        window.setWindowTitle(f"{self.config.entity_label}: создание записи")

    def make_ui(self) -> None:
        self.control_box = QtWidgets.QHBoxLayout()
        self.control_box.setContentsMargins(10, 5, 5, 5)
        self.form_box = QtWidgets.QVBoxLayout()

        self.create_button = make_button("Создать", icon=QtGui.QIcon(":/icons/new_record"))
        self.create_button.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)

        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        self.control_box.addWidget(self.create_button)
        self.control_box.addStretch()

        self.form_box.addWidget(self.form_body.draw())
        self.form_box.addStretch()

        self.layout.addLayout(self.control_box)
        self.layout.addLayout(self.form_box)

    @QtCore.pyqtSlot()
    def on_create_button_clicked(self):
        value = self.form_body.get_value()

        err = self.config.validate_record(value)
        if err is not None:
            make_warning(self, err, "Ошибка").exec()

            return

        try:
            id_ = self.config.repository.insert(self.config.to_database(value))
            self.record_created.emit(id_, value)

            make_info(self, "Запись создана", "Успешно").exec()
            self.record_creation_finalized.emit()
        except exc.IntegrityError as e:
            make_warning(self, self.config.process_commit_error(str(e.orig)), "Ошибка").exec()


class BasicRecordManager(RecordManager, QtWidgets.QWidget):
    record_edited = QtCore.pyqtSignal(dict)
    record_edition_finalized = QtCore.pyqtSignal()

    control_box: QtWidgets.QHBoxLayout
    form_box: QtWidgets.QVBoxLayout
    edit_button: QtWidgets.QPushButton
    layout: QtWidgets.QVBoxLayout

    form_body: FormBody
    record_id: int

    def __init__(self, window: WindowWidget, config: EntityConfig, id_: int,
                 initial_data: dict = None) -> None:
        super().__init__(window, config, id_, initial_data)

        self.config = config
        self.record_id = id_
        self.form_body = self.config.make_form(FormType.MANAGE)

        if initial_data is None:
            initial_data = self.config.from_database(self.config.repository.fetch_one(id_))
        self.form_body.set_value(initial_data)

        self.make_ui()

        self.edit_button.clicked.connect(self.on_save_button_clicked)

        window.setWindowTitle(f"{self.config.entity_label}: управление записью")

    def make_ui(self) -> None:
        self.control_box = QtWidgets.QHBoxLayout()
        self.control_box.setContentsMargins(10, 5, 5, 5)
        self.form_box = QtWidgets.QVBoxLayout()
        self.edit_button = make_button("Сохранить", icon=QtGui.QIcon(":/icons/save_record"))
        self.edit_button.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)

        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        self.control_box.addWidget(self.edit_button)
        self.control_box.addStretch()

        self.form_box.addWidget(self.form_body.draw())
        self.form_box.addStretch()

        self.layout.addLayout(self.control_box)
        self.layout.addLayout(self.form_box)

    @QtCore.pyqtSlot()
    def on_save_button_clicked(self) -> None:
        value = self.form_body.get_value()

        err = self.config.validate_record(value)
        if err is not None:
            make_warning(self, err, "Ошибка").exec()

            return

        try:
            self.config.repository.update(self.record_id, self.config.to_database(value))
            self.record_edited.emit(value)

            make_info(self, "Запись изменена", "Успешно").exec()
            self.record_edition_finalized.emit()
        except exc.IntegrityError as e:
            make_warning(self, self.config.process_commit_error(str(e.orig)), "Ошибка").exec()


class BasicRecordViewer(RecordViewer, QtWidgets.QWidget):
    control_box: QtWidgets.QHBoxLayout
    form_box: QtWidgets.QVBoxLayout
    close_button: QtWidgets.QPushButton
    layout: QtWidgets.QVBoxLayout

    form_body: FormBody

    def __init__(self, window: WindowWidget, config: EntityConfig, close_viewer: callable, id_: int,
                 initial_data: dict = None) -> None:
        super().__init__(window, config, close_viewer, id_, initial_data)

        self.config = config
        self.close_viewer = close_viewer
        self.form_body = self.config.make_form(FormType.VIEW)

        if initial_data is None:
            initial_data = self.config.from_database(self.config.repository.fetch_one(id_))
        self.form_body.set_value(initial_data)

        self.make_ui()

        self.close_button.clicked.connect(close_viewer)

        self.setWindowTitle(f"{self.config.entity_label}: просмотр записи")

    def make_ui(self) -> None:
        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        self.control_box = QtWidgets.QHBoxLayout()
        self.control_box.setContentsMargins(10, 5, 5, 5)
        self.form_box = QtWidgets.QVBoxLayout()
        self.close_button = make_button("Закрыть", icon=QtGui.QIcon(":/icons/close_record_view_screen"))
        self.close_button.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)

        self.control_box.addWidget(self.close_button)
        self.control_box.addStretch()

        self.form_box.addWidget(self.form_body.draw())
        self.form_box.addStretch()

        self.layout.addLayout(self.control_box)
        self.layout.addLayout(self.form_box)



from __future__ import annotations

from abc import abstractmethod
from typing import List, Dict, Optional, Type

from mado.data_repository import DataRepository
from mado.entity.operations import BasicRecordViewer, BasicRecordCreator, BasicRecordManager, RecordManager, \
    RecordCreator, \
    RecordViewer
from mado.form_body import FormBody, FormType


class EntityConfig:
    entity_label: str = "Наименование сущности"
    repository: DataRepository
    table_labels: List[str] = ["Наименование"]
    table_columns: List[str] = ["name"]

    manager: Type[RecordManager] = BasicRecordManager
    creator: Type[RecordCreator] = BasicRecordCreator
    viewer: Type[RecordViewer] = BasicRecordViewer

    def validate_record(self, data: Dict[str, any]) -> Optional[str]:
        return None

    def to_database(self, data: Dict[str, any]) -> Dict[str, any]:
        return data

    def from_database(self, data: Dict[str, any]) -> Dict[str, any]:
        return data

    def process_commit_error(self, err: str) -> str:
        return "Неизвестная ошибка"

    def process_delete_error(self, err: str) -> str:
        if err.startswith("FOREIGN KEY constraint failed"):
            return "Невозможно удалить эту запись, так как от неё зависит запись другой сущности"

    @abstractmethod
    def make_form(self, type_: FormType) -> FormBody:
        pass


class EventConfig(EntityConfig):
    def pre_save(self, data: Dict[str, any]) -> str | None:
        pass

    @abstractmethod
    def make_form(self, type_: FormType) -> FormBody:
        pass



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mado.application_window import Window


class ReadonlyCapable:
    @abstractmethod
    def set_readonly(self) -> None:
        pass



from enum import Enum
from typing import Tuple, List, Dict, Type

from PyQt5 import QtWidgets, QtCore, QtGui

from mado.entity.config import EntityConfig
from mado.input_base import InputBase
from mado.widgets.misc import make_label, make_button
from mado.widgets.modal import Modal
from mado.widgets.tables import MRecordTable


class ComboInput(InputBase, QtWidgets.QComboBox):
    variant_selected = QtCore.pyqtSignal(object)

    def __init__(self, variants: List[Tuple[any, str]], no_initial_selection: bool = False,
                 variant_selected: QtCore.pyqtSlot = None) -> None:
        super().__init__()

        for user_data, text in variants:
            self.addItem(text, user_data)

        if no_initial_selection:
            self.setCurrentIndex(-1)

        self.currentIndexChanged.connect(lambda *args, **kwargs: self.variant_selected.emit(self.get_value()))

        if variant_selected is not None:
            self.variant_selected.connect(variant_selected)

    def get_value(self) -> any:
        return self.currentData()

    def set_value(self, value: any) -> None:
        self.setCurrentIndex(self.findData(value))


class EnumInput(ComboInput):
    enum: Type[Enum]

    def __init__(self, enum: Type[Enum], no_initial_selection: bool = False,
                 variant_selected: QtCore.pyqtSlot = None) -> None:
        self.enum = enum
        to_list = [(x.name, x.value) for x in self.enum]

        super().__init__(to_list, no_initial_selection=no_initial_selection, variant_selected=variant_selected)

    def get_value(self) -> any:
        key = self.currentData()
        if key is None:
            return None
        return self.enum[key]

    def set_value(self, value: any) -> None:
        self.setCurrentIndex(self.findData(value.name))


class RelationChooserScreen(QtWidgets.QWidget):
    config: EntityConfig

    record_selected = QtCore.pyqtSignal(int, str)
    record_label_edited = QtCore.pyqtSignal(int, str)

    layout: QtWidgets.QVBoxLayout
    control_box: QtWidgets.QHBoxLayout

    table: MRecordTable
    confirm_button: QtWidgets.QPushButton
    create_button: QtWidgets.QPushButton
    manage_button: QtWidgets.QPushButton

    def __init__(self, config: EntityConfig) -> None:
        super().__init__()

        self.config = config

        self.confirm_button = make_button("Выбрать", icon=QtGui.QIcon(":/icons/select_record"))
        self.manage_button = make_button("Открыть запись")
        self.manage_button.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)
        self.manage_button.setEnabled(False)
        self.create_button = make_button("Создать", icon=QtGui.QIcon(":/icons/new_record"))
        self.create_button.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)
        self.confirm_button.setEnabled(False)
        self.control_box = QtWidgets.QHBoxLayout()
        self.control_box.addWidget(self.create_button)
        self.control_box.addWidget(self.manage_button)
        self.control_box.addStretch()
        self.control_box.addWidget(self.confirm_button)

        self.table = MRecordTable()
        self.table.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.table)
        self.layout.addLayout(self.control_box)

        self.setLayout(self.layout)

        self.table.row_double_clicked.connect(self.on_submit)
        self.confirm_button.clicked.connect(self.on_submit)
        self.table.row_selected.connect(self.on_table_row_selected)
        self.table.model.modelReset.connect(self.on_table_model_reset)

        self.create_button.clicked.connect(self.on_create_button_clicked)
        self.manage_button.clicked.connect(self.on_manage_button_clicked)

        self.table.model.setHorizontalHeaderLabels([self.config.table_labels[self.config.table_columns.index("name")]])
        self.fill_table()

    @QtCore.pyqtSlot()
    def on_table_row_selected(self) -> None:
        self.confirm_button.setEnabled(True)
        self.manage_button.setEnabled(True)

    @QtCore.pyqtSlot()
    def on_table_model_reset(self) -> None:
        self.confirm_button.setEnabled(False)
        self.manage_button.setEnabled(False)

    def fill_table(self) -> None:
        self.table.set_data(
            [([x.name], (x.id, x.name)) for x in
             self.config.repository.fetch_all()]
        )
        self.table.normalize()

    @QtCore.pyqtSlot()
    def on_submit(self) -> None:
        row = self.table.get_selected_row()
        if row is None:
            return
        self.record_selected.emit(*row)

    @QtCore.pyqtSlot()
    def on_create_button_clicked(self) -> None:
        modal = Modal(self)

        creator = self.config.creator(modal, self.config)

        created_id = -1
        created_data = dict()

        @QtCore.pyqtSlot(int, dict)
        def on_record_created(id_: int, data: Dict[str, any]) -> None:
            nonlocal created_id, created_data

            created_id = id_
            created_data = data

            self.fill_table()

        @QtCore.pyqtSlot()
        def on_record_creation_finalized() -> None:
            manager = self.config.manager(modal, self.config, created_id, created_data)
            manager.record_edited.connect(lambda *args, **kwargs: self.fill_table())

            modal.push_anon_screen(manager)

        creator.record_created.connect(on_record_created)
        creator.record_creation_finalized.connect(on_record_creation_finalized)

        modal.change_anon_screen(creator)
        modal.exec()

    @QtCore.pyqtSlot()
    def on_manage_button_clicked(self) -> None:
        modal = Modal(self)

        row = self.table.get_selected_row()
        if row is None:
            return

        id_ = row[0]

        manager = self.config.manager(modal, self.config, id_)

        @QtCore.pyqtSlot(dict)
        def record_edited(data: Dict[str, any]) -> None:
            self.fill_table()

            self.record_label_edited.emit(id_, data["name"])

        manager.record_edited.connect(record_edited)

        modal.change_anon_screen(manager)
        modal.exec()


class RelationInput(InputBase, QtWidgets.QWidget):
    layout: QtWidgets.QHBoxLayout
    record_label: QtWidgets.QLabel
    choose_button: QtWidgets.QPushButton

    chosen_id: int

    config: EntityConfig

    def __init__(self, config: EntityConfig) -> None:
        super().__init__()

        self.chosen_id = -1
        self.config = config
        self.title = f"{self.config.entity_label}: выбор связанной записи"

        self.record_label = make_label("<не выбрано>")
        self.record_label.setSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed)
        self.choose_button = make_button("Выбрать")
        self.choose_button.setSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        self.layout = QtWidgets.QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.record_label)
        self.layout.addWidget(self.choose_button)

        self.setLayout(self.layout)

        self.choose_button.clicked.connect(self.on_choose_button_clicked)

    @QtCore.pyqtSlot()
    def on_choose_button_clicked(self) -> None:
        modal = Modal(self)

        chooser = RelationChooserScreen(self.config)

        @QtCore.pyqtSlot(tuple)
        def choose_submitted(id_: int, label: str) -> None:
            self.chosen_id = id_
            self.record_label.setText(label)

            modal.accept()

        @QtCore.pyqtSlot(int, str)
        def record_label_edited(id_: int, label: str) -> None:
            if id_ == self.chosen_id:
                self.record_label.setText(label)

        chooser.record_selected.connect(choose_submitted)
        chooser.record_label_edited.connect(record_label_edited)

        size = self.window().screen().size()
        modal.resize(int(size.width() * 0.3), int(size.height() * 0.5))

        modal.change_anon_screen(chooser)
        modal.exec()

    def get_value(self) -> int:
        return self.chosen_id

    def set_value(self, value: int) -> None:
        self.chosen_id = value

        self.record_label.setText(self.config.repository.fetch_one(self.chosen_id).name)

    def set_readonly(self) -> None:
        self.choose_button.hide()



from __future__ import annotations

from abc import abstractmethod
from functools import singledispatchmethod
from typing import Dict, Callable, Type, TYPE_CHECKING

from PyQt5 import QtWidgets, QtCore

if TYPE_CHECKING:
    from mado.screen import ScreenWidget


class Window:
    @abstractmethod
    def register_screen(self, id_: str,
                        screen: Type[ScreenWidget] | Callable[[WindowWidget, ...], ScreenWidget]) -> None:
        pass

    @abstractmethod
    def change_screen(self, id_: str, *args, **kwargs) -> None:
        pass

    @abstractmethod
    def push_screen(self, id_: str, *args, **kwargs) -> None:
        pass

    @QtCore.pyqtSlot()
    def pop_screen(self) -> None:
        pass

    @abstractmethod
    def push_anon_screen(self, screen: ScreenWidget) -> None:
        pass

    @abstractmethod
    def change_anon_screen(self, screen: ScreenWidget) -> None:
        pass


class ApplicationWindow(Window, QtWidgets.QMainWindow):
    screens: Dict[str, Type[ScreenWidget] | Callable[[WindowWidget, ...], ScreenWidget]]
    stack: QtWidgets.QStackedLayout

    def __init__(self) -> None:
        super().__init__()

        self.stack = QtWidgets.QStackedLayout()
        self.screens = dict()

        self.resize(self.screen().size() * 0.83)

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(10, 10, 10, 10)

        layout.addLayout(self.stack)

        app_widget = QtWidgets.QWidget()
        app_widget.setLayout(layout)

        self.setCentralWidget(app_widget)

    def register_screen(self, id_: str,
                        screen: Type[ScreenWidget] | Callable[[WindowWidget, ...], ScreenWidget]) -> None:
        self.screens[id_] = screen

    def change_anon_screen(self, screen: ScreenWidget) -> None:
        self.push_anon_screen(screen)

        while self.stack.count() > 1:
            widget = self.stack.takeAt(0).widget()
            if widget:
                widget.deleteLater()

    def push_anon_screen(self, screen: ScreenWidget) -> None:
        self.stack.addWidget(screen)

        self.stack.setCurrentIndex(self.stack.count() - 1)

    def change_screen(self, id_: str, *args, **kwargs) -> None:
        self.change_anon_screen(self.screens[id_](self, *args, **kwargs))

    def push_screen(self, id_: str, *args, **kwargs) -> None:
        self.push_anon_screen(self.screens[id_](self, *args, **kwargs))

    @QtCore.pyqtSlot()
    def pop_screen(self) -> None:
        widget = self.stack.widget(self.stack.count() - 1)
        self.stack.removeWidget(widget)
        widget.deleteLater()

        screen: ScreenWidget = self.stack.widget(self.stack.count() - 1)
        screen.reconfigure()


WindowWidget = Window | QtWidgets.QWidget



from PyQt5 import QtWidgets

from mado.application_window import WindowWidget
from mado.form_body import FormBody
from mado.widgets.misc import make_back_button, make_label, make_font, make_button
from mado.widgets.tables import MReportTable


class Report(QtWidgets.QWidget):
    back_button: QtWidgets.QPushButton
    header_label: QtWidgets.QLabel

    top_box: QtWidgets.QHBoxLayout
    req_box: QtWidgets.QVBoxLayout
    layout: QtWidgets.QVBoxLayout

    table: MReportTable
    form_body: FormBody

    make_button_box: QtWidgets.QHBoxLayout
    make_button: QtWidgets.QPushButton

    def __init__(self, window: WindowWidget, form: FormBody) -> None:
        super().__init__()

        self.form_body = form
        self.form_body.draw()

        self.make_ui(window)

        self.header_label.setText("Отчёт такой-то")

    def make_ui(self, window: WindowWidget) -> None:
        self.back_button = make_back_button(window)
        self.header_label = make_label(font=make_font(size=14))
        self.top_box = QtWidgets.QHBoxLayout()
        self.top_box.addWidget(self.back_button)
        self.top_box.addWidget(self.header_label)
        self.top_box.addStretch()

        self.make_button = make_button("Составить отчёт")
        self.make_button_box = QtWidgets.QHBoxLayout()
        self.make_button_box.addStretch()
        self.make_button_box.addWidget(self.make_button)

        self.table = MReportTable()
        self.req_box = QtWidgets.QVBoxLayout()
        self.req_box.addWidget(self.form_body)
        self.req_box.addLayout(self.make_button_box)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addLayout(self.top_box)
        self.layout.addLayout(self.req_box)
        self.layout.addWidget(self.table)

        self.setLayout(self.layout)



from __future__ import annotations
import sys
import os
from typing import Tuple, Callable, TYPE_CHECKING, Dict

from PyQt5 import QtWidgets
from sqlalchemy import RowMapping

if TYPE_CHECKING:
    from mado.application_window import ApplicationWindow

app_name = ""
app: ApplicationWindow = None


def set_app(a: ApplicationWindow) -> None:
    global app

    app = a


def get_app() -> ApplicationWindow:
    return app


def set_app_name(name: str) -> None:
    global app_name

    app_name = name


def absolute_bundled_path(path: str) -> str:
    prefix = "."

    if getattr(sys, "frozen", False):
        prefix = sys._MEIPASS

    return os.path.join(prefix, path)


def absolute_appdata_path(path: str) -> str:
    prefix = "."

    if getattr(sys, "frozen", False):
        prefix = os.getenv("APPDATA")

    prefix = os.path.join(prefix, app_name)

    if not os.path.exists(prefix):
        os.makedirs(prefix)

    return os.path.join(prefix, path)


def make_warning(parent: any, msg: str, title: str = None) -> QtWidgets.QMessageBox:
    box = QtWidgets.QMessageBox(parent)

    box.setText(msg)
    box.setIcon(QtWidgets.QMessageBox.Icon.Warning)

    if title is not None:
        box.setWindowTitle(title)

    return box


def make_info(parent: any, msg: str, title: str = None) -> QtWidgets.QMessageBox:
    box = QtWidgets.QMessageBox(parent)

    box.setText(msg)
    box.setIcon(QtWidgets.QMessageBox.Icon.Information)

    if title is not None:
        box.setWindowTitle(title)

    return box


def row_mapping_to_dict(row: RowMapping) -> Dict[str, any]:
    return {key: value for key, value in row.items()}



from distutils.core import setup

from setuptools import find_packages

setup(
    name="madotsuki",
    version="1.2",
    description="Библиотека для построения графических CRUD-интерфейсов",
    author="jqmxaec",
    packages=find_packages(),
    install_requires=["PyQt5", "SQLAlchemy"],
)



from __future__ import annotations

from abc import abstractmethod

from mado.capabilities import ReadonlyCapable


class InputBase(ReadonlyCapable):
    @abstractmethod
    def get_value(self) -> any:
        raise NotImplementedError

    @abstractmethod
    def set_value(self, value: any):
        raise NotImplementedError

    def set_readonly(self) -> None:
        pass



from __future__ import annotations

from PyQt5 import QtWidgets, QtCore

from mado.application_window import WindowWidget
from mado.screen import Screen
from mado.widgets.misc import make_label, make_font, make_back_button, make_button


class MenuScreen(Screen, QtWidgets.QWidget):
    top_box: QtWidgets.QHBoxLayout
    layout: QtWidgets.QVBoxLayout
    header_label: QtWidgets.QLabel
    back_button: QtWidgets.QPushButton
    sections: QtWidgets.QGridLayout

    window_title: str
    window_instance: WindowWidget

    def __init__(self, window: WindowWidget, header: str = "", window_title: str = "",
                 need_back_button: bool = True) -> None:
        super().__init__()

        self.i = 0
        self.window_title = window_title
        self.window_instance = window

        self.window_instance.setWindowTitle(self.window_title)

        self.make_ui(self.window_instance, need_back_button)

        self.header_label.setText(header)

    def make_ui(self, window: WindowWidget, need_back_button: bool) -> None:
        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        self.header_label = make_label(font=make_font(size=14))
        self.top_box = QtWidgets.QHBoxLayout()

        if need_back_button:
            self.back_button = make_back_button(window)
            self.top_box.addWidget(self.back_button)

        self.top_box.addSpacing(5)
        self.top_box.addWidget(self.header_label)
        self.top_box.addStretch()

        self.sections = QtWidgets.QGridLayout()
        self.sections.setContentsMargins(20, 20, 20, 20)
        self.sections.setSpacing(25)

        self.layout.addLayout(self.top_box)
        self.layout.addLayout(self.sections)
        self.layout.addStretch()

    def add_section(self, section: MenuScreenSection):
        self.sections.addWidget(section, self.i // 2, self.i % 2)
        self.i += 1

    def reconfigure(self) -> None:
        self.window_instance.setWindowTitle(self.window_title)


class MenuScreenSection(QtWidgets.QWidget):
    header_label: QtWidgets.QLabel
    layout: QtWidgets.QVBoxLayout
    buttons: QtWidgets.QVBoxLayout

    def __init__(self, header: str = "") -> None:
        super().__init__()

        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        self.header_label = make_label(header, font=make_font(size=16, bold=True))
        self.header_label.setSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        self.header_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter)

        self.buttons = QtWidgets.QVBoxLayout()

        self.layout.addWidget(self.header_label)
        self.layout.addSpacing(7)
        self.layout.addLayout(self.buttons)
        self.layout.addStretch()

    def add_button(self, text: str = "", button_clicked: QtCore.pyqtSlot = None) -> None:
        button = make_button(text, font=make_font(underline=True), flat=True)
        button.setStyleSheet('color: "darkblue";')
        button.setSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        button.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)

        if button_clicked is not None:
            button.clicked.connect(button_clicked)

        self.buttons.addWidget(button)

    def add_existing_button(self, button: QtWidgets.QPushButton) -> None:
        self.buttons.addWidget(button)
